from __future__ import absolute_import, division, print_function

import re
import sys

from fast_dp.cell_spacegroup import constrain_cell, lattice_to_spacegroup

from fast_dp.pointless_reader import read_pointless_xml

def read_xds_idxref_lp(idxref_lp_file):
    '''Read the XDS IDXREF.LP file and return a dictionary indexed by the
    spacegroup number (ASSERT: this is the lowest symmetry spacegroup for
    the corresponding lattice) containing unit cell constants and a
    penalty. N.B. this also works from CORRECT.LP for the autoindexing
    results.'''

    # try doing this with regular expression: * int lattice...`

    regexp = re.compile(r'^ \*\ ')

    results = { }

    for record in open(idxref_lp_file, 'r').readlines():
        if regexp.match(record):
            tokens = record.split()
            spacegroup = lattice_to_spacegroup(tokens[2])
            cell = tuple(map(float, tokens[4:10]))
            constrained_cell = constrain_cell(tokens[2][0], cell)
            penalty = float(tokens[3])

            if spacegroup in results:
                if penalty < results[spacegroup][0]:
                    results[spacegroup] = penalty, constrained_cell
            else:
                results[spacegroup] = penalty, constrained_cell

        if 'DETECTOR COORDINATES (PIXELS) OF DIRECT BEAM' in record:
            results['beam centre pixels'] = map(float, record.split()[-2:])

    assert('beam centre pixels' in results)

    return results

def read_xds_correct_lp(correct_lp_file):
    '''Read the XDS CORRECT.LP file and get out the spacegroup and
    unit cell constants it decided on.'''

    unit_cell = None
    space_group_number = None

    for record in open(correct_lp_file, 'r').readlines():
        if 'SPACE_GROUP_NUMBER=' in record:
            try:
                space_group_number = int(record.split()[-1])
            except:
                space_group_number = 0
        if 'UNIT_CELL_CONSTANTS=' in record and not 'used' in record:
            unit_cell = tuple(map(float, record.split()[-6:]))

    return unit_cell, space_group_number

def read_correct_lp_get_resolution(correct_lp_file):
    '''Read the CORRECT.LP file and get an estimate of the resolution limit.
    This should then be recycled to a rerun of CORRECT, from which the
    reflections will be merged to get the statistics.'''

    correct_lp = open(correct_lp_file, 'r').readlines()

    rec = -1

    for j in range(len(correct_lp)):
        record = correct_lp[j]

        if 'RESOLUTION RANGE  I/Sigma  Chi^2  R-FACTOR  R-FACTOR' in record:
            rec = j + 3
            break

    if rec < 0:
        raise RuntimeError('resolution information not found')

    j = rec

    while not '--------' in correct_lp[j]:
        isigma = float(correct_lp[j].split()[2])
        if isigma < 1:
            return float(correct_lp[j].split()[1])
        j += 1

    # this will assume that strong reflections go to the edge of the detector
    # => do not need to feed back a resolution limit...

    return None



from __future__ import absolute_import, division, print_function

import os
import shutil

from fast_dp.xds_writer import write_xds_inp_correct
from fast_dp.run_job import run_job
from fast_dp.cell_spacegroup import spacegroup_number_to_name

def scale(unit_cell, metadata, space_group_number, resolution_high):
    '''Perform the scaling with the spacegroup and unit cell calculated
    from pointless and correct. N.B. this scaling is done by CORRECT.'''

    assert(unit_cell)
    assert(metadata)
    assert(space_group_number)

    if resolution_high:
        resolution_high = resolution_high
    else:
        resolution_high = 0.0

    xds_inp = 'CORRECT.INP'

    write_xds_inp_correct(metadata, unit_cell,
                          space_group_number, xds_inp,
                          resolution_high = resolution_high)

    shutil.copyfile(xds_inp, 'XDS.INP')

    run_job('xds_par')

    # once again should check on the general happiness of everything...

    for step in ['CORRECT']:
        lastrecord = open('%s.LP' % step).readlines()[-1]
        if '!!! ERROR !!!' in lastrecord:
            raise RuntimeError('error in %s: %s' % \
                  (step, lastrecord.replace('!!! ERROR !!!', '').strip()))

    # and get the postrefined cell constants from GXPARM.XDS - but continue
    # to work for the old format too...

    if 'XPARM.XDS' in open('GXPARM.XDS', 'r').readline():
        # new format
        gxparm = open('GXPARM.XDS', 'r').readlines()
        space_group = spacegroup_number_to_name(int(gxparm[3].split()[0]))
        unit_cell = tuple(map(float, gxparm[3].split()[1:]))
    else:
        # old format:
        gxparm = open('GXPARM.XDS', 'r').readlines()
        space_group = spacegroup_number_to_name(int(gxparm[7].split()[0]))
        unit_cell = tuple(map(float, gxparm[7].split()[1:]))

    # FIXME also get the postrefined mosaic spread out...

    space_group = space_group
    unit_cell = unit_cell

    # and the total number of good reflections
    nref = 0

    # set the refined beam in case it was not refined correctly in the
    # global refinement (this probably indicates a bigger problem anyway)
    refined_beam = 0, 0

    for record in open('CORRECT.LP', 'r').readlines():
        if 'NUMBER OF ACCEPTED OBSERVATIONS' in record:
            nref = int(record.replace(')', ') ').split()[-1])
        if 'DETECTOR COORDINATES (PIXELS) OF DIRECT BEAM' in record:
            refined_beam = tuple(map(float, record.split()[-2:]))

    # hack in xdsstat (but don't cry if it fails)

    try:
        xdsstat_output = run_job('xdsstat', [], ['XDS_ASCII.HKL'])
        open('xdsstat.log', 'w').write(''.join(xdsstat_output))
    except:
        pass

    return unit_cell, space_group, nref, refined_beam



from __future__ import absolute_import, division, print_function

import os
import shutil

from fast_dp.xds_writer import write_xds_inp_integrate
from fast_dp.run_job import run_job

def integrate(metadata, p1_unit_cell, resolution_low, n_jobs, n_processors):
    '''Peform the integration with a triclinic basis.'''

    assert(metadata)
    assert(p1_unit_cell)

    xds_inp = 'INTEGRATE.INP'

    # FIXME in here make a calculation from the metadata of a sensible
    # maximum number of jobs, to give minimally 5 degree wedges.

    write_xds_inp_integrate(metadata, xds_inp, resolution_low,
                            no_jobs=n_jobs, no_processors=n_processors)

    shutil.copyfile(xds_inp, 'XDS.INP')

    run_job('xds_par')

    # FIXME need to check that all was hunky-dory in here!

    for step in ['DEFPIX', 'INTEGRATE']:
        if not os.path.exists('%s.LP' % step):
            continue
        lastrecord = open('%s.LP' % step).readlines()[-1]
        if '!!! ERROR !!!' in lastrecord:
            raise RuntimeError('error in %s: %s' % \
                  (step, lastrecord.replace(
                '!!! ERROR !!!', '').strip().lower()))

    if not os.path.exists('INTEGRATE.LP'):
        step = 'INTEGRATE'
        for record in open('LP_01.tmp').readlines():
            if '!!! ERROR !!! AUTOMATIC DETERMINATION OF SPOT SIZE ' in record:
                raise RuntimeError('error in %s: %s' % \
                      (step, record.replace(
                    '!!! ERROR !!!', '').strip().lower()))
            elif '!!! ERROR !!! CANNOT OPEN OR READ FILE LP_01.tmp' in record:
                raise RuntimeError('integration error: cluster error')

    # check for some specific errors

    for step in ['INTEGRATE']:
        for record in open('%s.LP' % step).readlines():
            if '!!! ERROR !!! AUTOMATIC DETERMINATION OF SPOT SIZE ' in record:
                raise RuntimeError('error in %s: %s' % \
                      (step, record.replace(
                    '!!! ERROR !!!', '').strip().lower()))
            elif '!!! ERROR !!! CANNOT OPEN OR READ FILE LP_01.tmp' in record:
                raise RuntimeError('integration error: cluster error')



    # if all was ok, look in the working directory for files named
    # forkintegrate_job.o341858 &c. and remove them. - N.B. this is site
    # specific!

    for f in os.listdir(os.getcwd()):
        if 'forkintegrate_job.' in f[:18]:
            try:
                os.remove(f)
            except:
                pass

    # get the mosaic spread ranges

    mosaics = []

    for record in open('INTEGRATE.LP'):
        if 'CRYSTAL MOSAICITY (DEGREES)' in record:
            mosaics.append(float(record.split()[-1]))

    mosaic = sum(mosaics) / len(mosaics)

    return min(mosaics), mosaic, max(mosaics)



from __future__ import absolute_import, division, print_function

import os

from cctbx import xray
from cctbx.sgtbx import space_group
from cctbx.sgtbx import space_group_symbols
from cctbx.uctbx import unit_cell
from cctbx.crystal import symmetry

def ersatz_pointgroup(spacegroup_name):
    '''Guess the pointgroup for the spacegroup by mapping from short to
    long name, then taking 1st character from each block.'''

    pg = None

    for record in open(
        os.path.join(os.environ['CLIBD'], 'symop.lib'), 'r').readlines():
        if ' ' in record[:1]:
            continue
        if spacegroup_name == record.split()[3]:
            pg = record.split()[4][2:]
        elif spacegroup_name == record.split('\'')[1].replace(' ', ''):
            pg = record.split()[4][2:]

    if not pg:
        raise RuntimeError('spacegroup %s unknown' % spacegroup_name)

    # FIXME this is probably not correct for small molecule work...
    # just be aware of this, in no danger right now of handling non-chiral
    # spacegroups

    if '/' in pg:
        pg = pg.split('/')[0]

    result = spacegroup_name[0] + pg

    if 'H3' in result:
        result = result.replace('H3', 'R3')

    return result

def spacegroup_to_lattice(input_spacegroup):
    ''' This generates a lattics from a the imported file bu chopping off
    the first letter of the cell type, changing to lowercase and then
    prepending it to the first letter of the spacegroup.'''

    def fix_hH(lattice):
        if lattice != 'hH':
            return lattice
        return 'hR'

    mapping = {'TRICLINIC':'a',
               'MONOCLINIC':'m',
               'ORTHORHOMBIC':'o',
               'TETRAGONAL':'t',
               'TRIGONAL':'h',
               'HEXAGONAL':'h',
               'CUBIC':'c'}

    if type(input_spacegroup) == type(u''):
        input_spacegroup = str(input_spacegroup)

    if type(input_spacegroup) == type(''):
        for record in open(
            os.path.join(os.environ['CLIBD'], 'symop.lib'), 'r').readlines():
            if ' ' in record[:1]:
                continue
            if input_spacegroup == record.split()[3]:
                return fix_hH(mapping[record.split()[5]] + record.split()[3][0])

    elif type(input_spacegroup) == type(0):
        for record in open(
            os.path.join(os.environ['CLIBD'], 'symop.lib'), 'r').readlines():
            if ' ' in record[:1]:
                continue
            if input_spacegroup == int(record.split()[0]):
                return fix_hH(mapping[record.split()[5]] + record.split()[3][0])

    else:
        raise RuntimeError('bad type for input: %s' % type(input_spacegroup))

    return None

def check_spacegroup_name(spacegroup_name):
    '''Will return normalised name if spacegroup name is recognised,
    raise exception otherwise. For checking command-line options.'''

    try:
        j = int(spacegroup_name)
        if j > 230 or j <= 0:
            raise RuntimeError('spacegroup number nonsense: %s' \
                  % spacegroup_name)
        return spacegroup_number_to_name(j)

    except ValueError:
        pass

    found_spacegroup = None

    for record in open(
        os.path.join(os.environ['CLIBD'], 'symop.lib'), 'r').readlines():
        if ' ' in record[:1]:
            continue
        if spacegroup_name == record.split()[3]:
            return spacegroup_name

    raise RuntimeError('spacegroup name "%s" not recognised' % spacegroup_name)

def check_split_cell(cell_string):
    '''Will return tuple of floats a, b, c, alpha, beta, gamma from input
    cell string which contains a,b,c,alpha,beta,gamma raising an exception
    if there is a problem.'''

    ideal_string = 'a,b,c,alpha,beta,gamma'

    if not cell_string.count(',') == 5:
        raise RuntimeError('%s should be of the form %s' % \
              (cell_string, ideal_string))

    a, b, c, alpha, beta, gamma = tuple(
        map(float, cell_string.split(',')))

    return a, b, c, alpha, beta, gamma

def constrain_cell(lattice_class, cell):
    '''Constrain cell to fit lattice class x.'''

    a, b, c, alpha, beta, gamma = cell

    if lattice_class == 'a':
        return (a, b, c, alpha, beta, gamma)
    elif lattice_class == 'm':
        return (a, b, c, 90.0, beta, 90.0)
    elif lattice_class == 'o':
        return (a, b, c, 90.0, 90.0, 90.0)
    elif lattice_class == 't':
        e = (a + b) / 2.0
        return (e, e, c, 90.0, 90.0, 90.0)
    elif lattice_class == 'h':
        e = (a + b) / 2.0
        return (e, e, c, 90.0, 90.0, 120.0)
    elif lattice_class == 'c':
        e = (a + b + c) / 3.0
        return (e, e, e, 90.0, 90.0, 90.0)

    raise RuntimeError('lattice class not recognised: %s' % lattice_class)

def spacegroup_number_to_name(spg_num):
    '''Convert a spacegroup number to a more readable name.'''

    database = { }

    for record in open(
        os.path.join(os.environ['CLIBD'], 'symop.lib'), 'r').readlines():
        if ' ' in record[:1]:
            continue
        number = int(record.split()[0])
        name = record.split('\'')[1].strip()
        database[number] = name

    return database[spg_num]

def lattice_to_spacegroup(lattice):
    ''' Converts a lattice to the spacegroup with the lowest symmetry
    possible for that lattice'''
    l2s = {
        'aP':1,   'mP':3,   'mC':5,   'mI':5,
        'oP':16,  'oC':21,  'oI':23,  'oF':22,
        'tP':75,  'tI':79,  'hP':143, 'hR':146,
        'hH':146, 'cP':195, 'cF':196, 'cI':197
        }

    return l2s[lattice]

def lauegroup_to_lattice(lauegroup):
    '''Convert a Laue group representation (from pointless, e.g. I m m m)
    to something useful, like the implied crystal lattice (in this
    case, oI.)'''

    # this has been calculated from the results of Ralf GK's sginfo and a
    # little fiddling...
    #
    # 19/feb/08 added mI record as pointless has started producing this -
    # why??? this is not a "real" spacegroup... may be able to switch this
    # off...
    #                             'I2/m': 'mI',

    lauegroup_to_lattice = {'Ammm': 'oA',
                            'C2/m': 'mC',
                            'I2/m': 'mI',
                            'Cmmm': 'oC',
                            'Fm-3': 'cF',
                            'Fm-3m': 'cF',
                            'Fmmm': 'oF',
                            'H-3': 'hR',
                            'H-3m': 'hR',
                            'R-3:H': 'hR',
                            'R-3m:H': 'hR',
                            'R-3': 'hR',
                            'R-3m': 'hR',
                            'I4/m': 'tI',
                            'I4/mmm': 'tI',
                            'Im-3': 'cI',
                            'Im-3m': 'cI',
                            'Immm': 'oI',
                            'P-1': 'aP',
                            'P-3': 'hP',
                            'P-3m': 'hP',
                            'P2/m': 'mP',
                            'P4/m': 'tP',
                            'P4/mmm': 'tP',
                            'P6/m': 'hP',
                            'P6/mmm': 'hP',
                            'Pm-3': 'cP',
                            'Pm-3m': 'cP',
                            'Pmmm': 'oP'}

    updated_laue = ''

    for l in lauegroup.split():
        if not l == '1':
            updated_laue += l

    return lauegroup_to_lattice[updated_laue]

def generate_primitive_cell(unit_cell_constants, space_group_name):
    '''For a given set of unit cell constants and space group, determine the
    corresponding primitive unit cell...'''

    uc = unit_cell(unit_cell_constants)
    sg = space_group(space_group_symbols(space_group_name).hall())
    cs = symmetry(unit_cell = uc,
                  space_group = sg)
    csp = cs.change_basis(cs.change_of_basis_op_to_primitive_setting())

    return csp.unit_cell()

if __name__ == '__main__':

    import sys

    for token in sys.argv[1:]:
        print(ersatz_pointgroup(token))



from __future__ import absolute_import, division, print_function

import os
import shutil

from fast_dp.xds_reader import read_xds_idxref_lp, read_correct_lp_get_resolution, \
     read_xds_correct_lp
from fast_dp.pointless_reader import read_pointless_xml
from fast_dp.xds_writer import write_xds_inp_correct_no_cell, write_xds_inp_correct
from fast_dp.run_job import run_job
from fast_dp.cell_spacegroup import lattice_to_spacegroup, ersatz_pointgroup, \
    spacegroup_to_lattice, check_spacegroup_name

from fast_dp.logger import write

def decide_pointgroup(p1_unit_cell, metadata,
                      input_spacegroup = None):
    '''Run POINTLESS to get the list of allowed pointgroups (N.B. will
    insist on triclinic symmetry for this scaling step) then run
    pointless on the resulting reflection file to get the idea of the
    best pointgroup to use. Then return the correct pointgroup and
    cell.'''

    assert(p1_unit_cell)

    # write correct input file for the triclinic solution, in the
    # working directory

    xds_inp = 'P1.INP'

    write_xds_inp_correct(metadata, p1_unit_cell, 1,
                          xds_inp, turn_subset=True)

    shutil.copyfile(xds_inp, 'XDS.INP')

    run_job('xds_par')

    shutil.copyfile('CORRECT.LP', 'P1.LP')

    # get the list of allowed lattices

    results = read_xds_idxref_lp('CORRECT.LP')

    # also read out the resolution limit

    resolution_high = read_correct_lp_get_resolution('CORRECT.LP')

    # run pointless, get the list of suggested lattices and pointgroups
    # FIXME should use the program manager for this... yes, this will
    # check that the executable is available too!

    xdsin = 'XDS_ASCII.HKL'
    xmlout = 'pointless.xml'

    pointless_log = run_job(
        'pointless',
        arguments = ['xdsin', xdsin, 'xmlout', xmlout],
        stdin = ['systematicabsences off'])

    fout = open('pointless.log', 'w')

    for record in pointless_log:
        fout.write(record)

    fout.close()

    # now read the XML file

    pointless_results = read_pointless_xml(xmlout)

    # select the top solution which is allowed, return this

    if input_spacegroup:
        sg_accepted = False;
        pointgroup = ersatz_pointgroup(input_spacegroup)
        if pointgroup.startswith('H'):
            pointgroup = pointgroup.replace('H', 'R')
        lattice = spacegroup_to_lattice(input_spacegroup)
        for r in pointless_results:
            result_sg = "".join(check_spacegroup_name(r[1]).split(' '))
            if lattice_to_spacegroup(lattice) in results and \
                   ersatz_pointgroup(result_sg) == pointgroup :
                space_group_number = r[1]
                unit_cell = results[lattice_to_spacegroup(r[0])][1]
                write('Happy with sg# %d' % space_group_number)
                write('%6.2f %6.2f %6.2f %6.2f %6.2f %6.2f' %
                      unit_cell)
                sg_accepted = True
                break

        if not sg_accepted:
            write('No indexing solution for spacegroup %s so ignoring' % \
                  input_spacegroup)
            input_spacegroup = None

    # if input space group obviously nonsense, allow to ignore just warn
    if not input_spacegroup:
        for r in pointless_results:
            if lattice_to_spacegroup(r[0]) in results:
                space_group_number = r[1]
                unit_cell = results[lattice_to_spacegroup(r[0])][1]
                write('Happy with sg# %d' % space_group_number)
                write('%6.2f %6.2f %6.2f %6.2f %6.2f %6.2f' %
                      unit_cell)

                break
            else:
                write('Rejected solution %s %3d' % r)

    # this should probably be a proper check...
    assert(space_group_number)

    # also save the P1 XDS_ASCII.HKL file see
    # http://trac.diamond.ac.uk/scientific_software/ticket/1106

    shutil.copyfile('XDS_ASCII.HKL', 'XDS_P1.HKL')

    return unit_cell, space_group_number, resolution_high



from __future__ import absolute_import, division, print_function

import os
import time

from fast_dp.logger import write

from fast_dp.image_names import image2template_directory, find_matching_images, \
    template_directory_number2image

from fast_dp.run_job import run_job

def check_file_readable(filename):
    '''Check that the file filename exists and that it can be read. Returns
    only if everything is OK.'''

    if not os.path.exists(filename):
        raise RuntimeError('file %s not found' % filename)

    if not os.access(filename, os.R_OK):
        raise RuntimeError('file %s not readable' % filename)

    return

def get_dectris_serial_no(record):
    if not 'S/N' in record:
        return '0'
    tokens = record.split()
    return tokens[tokens.index('S/N') + 1]

__hdf5_lib = ''
def find_hdf5_lib(lib_name=None):
  global __hdf5_lib
  if __hdf5_lib:
    return __hdf5_lib
  if lib_name is None:
    'dectris-neggia.so'
  for d in os.environ['PATH'].split(os.pathsep):
    if os.path.exists(os.path.join(d, 'xds_par')):
      __hdf5_lib = 'LIB=%s\n' % os.path.join(d, lib_name)
      return __hdf5_lib
  return ''

try:
  import bz2
except: # intentional
  bz2 = None

try:
  import gzip
except: # intentional
  gzip = None


def is_bz2(filename):
  if not '.bz2' in filename[-4:]:
    return False
  return 'BZh' in open(filename, 'rb').read(3)

def is_gzip(filename):
  if not '.gz' in filename[-3:]:
    return False
  magic = open(filename, 'rb').read(2)
  return ord(magic[0]) == 0x1f and ord(magic[1]) == 0x8b

def open_file(filename, mode='rb', url=False):
  if is_bz2(filename):
    if bz2 is None:
      raise RuntimeError('bz2 file provided without bz2 module')
    fh_func = lambda: bz2.BZ2File(filename, mode)

  elif is_gzip(filename):
    if gzip is None:
      raise RuntimeError('gz file provided without gzip module')
    fh_func = lambda: gzip.GzipFile(filename, mode)

  else:
    fh_func = lambda: open(filename, mode)

  return fh_func()

def failover_hdf5(hdf5_file, lib_name=None):
    from dxtbx.serialize import xds
    from dxtbx.datablock import DataBlockFactory
    t0 = time.time()
    db = DataBlockFactory.from_filenames([hdf5_file])[0]
    sweep = db.extract_sweeps()[0]
    t1 = time.time()
    write('Reading %s took %.2fs' % (hdf5_file, t1-t0))
    d = sweep.get_detector()
    s = sweep.get_scan()
    g = sweep.get_goniometer()
    b = sweep.get_beam()

    # returns slow, fast, convention here is reverse
    size = tuple(reversed(d[0].get_image_size()))

    size0k_to_class = {1:'eiger 1M',
                       2:'eiger 4M',
                       3:'eiger 9M',
                       4:'eiger 16M'}

    header = { }

    header['detector_class'] = size0k_to_class[int(size[0]/1000)]
    header['detector'] = size0k_to_class[int(size[0]/1000)].upper().replace(
        ' ', '_')
    header['size'] = size
    header['serial_number'] = 0
    header['extra_text'] = find_hdf5_lib(lib_name)

    header['phi_start'] = s.get_angle_from_image_index(1.0, deg=True)
    header['phi_end'] = s.get_angle_from_image_index(2.0, deg=True)
    header['phi_width'] = header['phi_end'] - header['phi_start']
    header['oscillation'] = header['phi_start'], header['phi_width']
    header['exposure_time'] = s.get_exposure_times()[0]
    header['oscillation_axis'] = 'Omega_I_guess'
    axis = g.get_rotation_axis()
    if abs(axis[1]) > abs(axis[0]):
        header['goniometer_is_vertical'] = True
    header['distance'] = d[0].get_distance()
    header['wavelength'] = b.get_wavelength()
    header['pixel'] = d[0].get_pixel_size()
    header['saturation'] = d[0].get_trusted_range()[1]
    header['sensor'] = d[0].get_thickness()
    header['beam'] = d[0].get_beam_centre(b.get_s0())
    images = s.get_image_range()
    directory, template = os.path.split(hdf5_file)
    header['directory'] = directory
    header['template'] = template.replace('master', '??????')
    header['start'] = images[0]
    header['end'] = images[1]
    header['matching'] = range(images[0], images[1]+1)
    return header

def failover_cbf(cbf_file):
    '''CBF files from the latest update to the PILATUS detector cause a
    segmentation fault in diffdump. This is a workaround.'''

    header = { }

    header['two_theta'] = 0.0

    for record in open_file(cbf_file):
        if '_array_data.data' in record:
            break

        if 'EIGER 1M' in record.upper():
            header['detector_class'] = 'eiger 1M'
            header['detector'] = 'dectris'
            header['size'] = (1065, 1030)
            header['serial_number'] = get_dectris_serial_no(record)
            continue

        if 'EIGER 4M' in record.upper():
            header['detector_class'] = 'eiger 4M'
            header['detector'] = 'dectris'
            header['size'] = (2176, 2070)
            header['serial_number'] = get_dectris_serial_no(record)
            continue

        if 'EIGER 9M' in record.upper():
            header['detector_class'] = 'eiger 9M'
            header['detector'] = 'dectris'
            header['size'] = (3269, 3110)
            header['serial_number'] = get_dectris_serial_no(record)
            continue

        if 'EIGER 16M' in record.upper():
            header['detector_class'] = 'eiger 16M'
            header['detector'] = 'dectris'
            header['size'] = (4371, 4150)
            header['serial_number'] = get_dectris_serial_no(record)
            continue

        if 'PILATUS 2M' in record:
            header['detector_class'] = 'pilatus 2M'
            header['detector'] = 'dectris'
            header['size'] = (1679, 1475)
            header['serial_number'] = get_dectris_serial_no(record)
            continue

        if 'PILATUS 6M' in record:
            header['detector_class'] = 'pilatus 6M'
            header['detector'] = 'dectris'
            header['size'] = (2527, 2463)
            header['serial_number'] = get_dectris_serial_no(record)
            continue

        if 'PILATUS3 6M' in record:
            header['detector_class'] = 'pilatus 6M'
            header['detector'] = 'dectris'
            header['size'] = (2527, 2463)
            header['serial_number'] = get_dectris_serial_no(record)
            continue

        if 'PILATUS 12M' in record:
            header['detector_class'] = 'pilatus 12M'
            header['detector'] = 'dectris'
            header['size'] = (5071, 2463)
            header['serial_number'] = get_dectris_serial_no(record)
            continue

        if 'Detector: ADSC HF-4M' in record:
            header['detector_class'] = 'adsc 4M'
            header['detector'] = 'adsc-pad'
            header['size'] = (2290, 2100)
            header['serial_number'] = record.replace(',', '').split()[-1]
            continue

        if 'Start_angle' in record:
            header['phi_start'] = float(record.split()[-2])
            continue

        if 'Angle_increment' in record:
            header['phi_width'] = float(record.split()[-2])
            header['phi_end'] = header['phi_start'] + header['phi_width']
            header['oscillation'] = header['phi_start'], header['phi_width']
            continue

        if 'Exposure_period' in record:
            header['exposure_time'] = float(record.split()[-2])
            continue

        if 'Detector_distance' in record:
            header['distance'] = 1000 * float(record.split()[2])
            continue

        if 'Oscillation_axis' in record:
            header['oscillation_axis'] = record.split('axis')[-1].strip()

        if 'Wavelength' in record:
            header['wavelength'] = float(record.split()[-2])
            continue

        if 'Pixel_size' in record:
            header['pixel'] = 1000 * float(record.split()[2]), \
                              1000 * float(record.split()[5])
            continue

        if 'Count_cutoff' in record:
            header['saturation'] = int(record.split()[2])

        if 'Silicon sensor' in record:
            header['sensor'] = 1000 * float(record.split()[4])

        if 'Beam_xy' in record:
            beam_pixels = map(float, record.replace('(', '').replace(
                ')', '').replace(',', '').split()[2:4])

            # for CBF images need to swap these to put in XDS mosflm
            # coordinate frame...
            header['beam'] = beam_pixels[0] * header['pixel'][0], \
                             beam_pixels[1] * header['pixel'][1]

            continue

        # try to get the date etc. literally.

        try:
            datestring = record.split()[-1].split('.')[0]
            format = '%Y-%b-%dT%H:%M:%S'
            struct_time = time.strptime(datestring, format)
            header['date'] = time.asctime(struct_time)
            header['epoch'] = time.mktime(struct_time)

        except:
            pass

        try:
            datestring = record.split()[-1].split('.')[0]
            format = '%Y-%m-%dT%H:%M:%S'
            struct_time = time.strptime(datestring, format)
            header['date'] = time.asctime(struct_time)
            header['epoch'] = time.mktime(struct_time)

        except:
            pass

        try:
            datestring = record.replace('#', '').strip().split('.')[0]
            format = '%Y/%b/%d %H:%M:%S'
            struct_time = time.strptime(datestring, format)
            header['date'] = time.asctime(struct_time)
            header['epoch'] = time.mktime(struct_time)

        except:
            pass

    # cope with vertical goniometer on I24 @ DLS from 2015/1/1
    # and possible alternative horizontal gonio if +FAST
    if header.get('serial_number', '0') == '60-0119' and \
            int(header['date'].split()[-1]) >= 2015:
        if '+FAST' in header.get('oscillation_axis', ''):
            header['goniometer_is_vertical'] = False
        else:
            header['goniometer_is_vertical'] = True

    else:
        if not 'goniometer_is_vertical' in header:
            header['goniometer_is_vertical'] = False

    return header

__lib_name = None
def set_lib_name(lib_name):
    global __lib_name
    __lib_name = lib_name

def read_image_metadata(image):
    '''Read the image header and send back the resulting metadata in a
    dictionary.'''

    assert(os.path.exists(image))

    if image.endswith('.h5'):
        assert 'master' in image
        global __lib_name

        metadata = failover_hdf5(image, lib_name=__lib_name)
        if metadata['goniometer_is_vertical']:
            metadata['detector'] = '%sV' % metadata['detector']

        return metadata

    # FIXME also check that the file can also be read - assert is acceptable,
    # also use the first image in the wedge to get the frame metadata

    template, directory = image2template_directory(image)

    matching = find_matching_images(template, directory)
    image = template_directory_number2image(template, directory, min(matching))

    # work around (preempt) diffdump failure with the new 2M instrument
    # FIXME may also need to do this for the new 6M instrument which is
    # incoming...

    try:
        if '.cbf' in image:
            metadata = failover_cbf(image)

            assert(metadata['detector_class'] in \
                   ['pilatus 2M', 'pilatus 6M', 'pilatus 12M',
                    'eiger 1M', 'eiger 4M', 'eiger 9M', 'eiger 16M',
                    'adsc 4M'])

            if metadata['detector_class'] == 'pilatus 2M':
                metadata['detector'] = 'PILATUS_2M'
            elif metadata['detector_class'] == 'pilatus 12M':
                metadata['detector'] = 'PILATUS_12M'
            elif metadata['detector_class'] == 'eiger 1M':
                metadata['detector'] = 'EIGER_1M'
            elif metadata['detector_class'] == 'eiger 4M':
                metadata['detector'] = 'EIGER_4M'
            elif metadata['detector_class'] == 'eiger 9M':
                metadata['detector'] = 'EIGER_9M'
            elif metadata['detector_class'] == 'eiger 16M':
                metadata['detector'] = 'EIGER_16M'
            elif metadata['detector_class'] == 'adsc 4M':
                metadata['detector'] = 'ADSC_4M'
            else:
                metadata['detector'] = 'PILATUS_6M'

            # handle I24 @ DLS vertical goniometer from 2015/1/1
            if metadata['goniometer_is_vertical']:
                metadata['detector'] = '%sV' % metadata['detector']

            metadata['directory'] = directory
            metadata['template'] = template
            metadata['start'] = min(matching)
            metadata['end'] = max(matching)

            metadata['matching'] = matching

            return metadata

    except ValueError:
        pass

    # MAR CCD images record the beam centre in pixels...

    diffdump_output = run_job('diffdump', arguments = [image])

    metadata = { }

    for record in diffdump_output:
        if 'Wavelength' in record:
            wavelength = float(record.split()[-2])
            metadata['wavelength'] = wavelength

        elif 'Beam center' in record:
            x = float(record.replace('(', ' ').replace(
                'mm', ' ').replace(',', ' ').split()[3])
            y = float(record.replace('(', ' ').replace(
                'mm', ' ').replace(',', ' ').split()[4])
            metadata['beam'] = x, y

        elif 'Image Size' in record:
            x = int(record.replace('(', ' ').replace(
                'px', ' ').replace(',', ' ').split()[3])
            y = int(record.replace('(', ' ').replace(
                'px', ' ').replace(',', ' ').split()[4])
            metadata['size'] = x, y

        elif 'Pixel Size' in record:
            x = float(record.replace('(', ' ').replace(
                'mm', ' ').replace(',', ' ').split()[3])
            y = float(record.replace('(', ' ').replace(
                'mm', ' ').replace(',', ' ').split()[4])
            metadata['pixel'] = x, y

        elif 'Distance' in record:
            distance = float(record.split()[-2])
            metadata['distance'] = distance

        elif 'Exposure time' in record:
            metadata['exposure_time'] = float(record.split()[-2])

        elif 'Oscillation' in record:
            phi_start = float(record.split()[3])
            phi_end = float(record.split()[5])
            phi_width = phi_end - phi_start

            if phi_width > 360.0:
                phi_width -= 360.0

            metadata['phi_start'] = phi_start
            metadata['phi_width'] = phi_width
            metadata['phi_end'] = phi_end
            metadata['oscillation'] = phi_start, phi_width

        elif 'Manufacturer' in record or 'Image type' in record:
            detector = record.split()[-1]
            if detector == 'ADSC':
                metadata['detector'] = 'ADSC'
            elif detector == 'MAR':
                metadata['detector'] = 'MARCCD'
            elif detector == 'DECTRIS':
                metadata['detector'] = 'PILATUS_6M'
            elif detector == 'RIGAKU':
                metadata['detector'] = 'RIGAKU'
            else:
                raise RuntimeError('detector %s not yet supported' % \
                      detector)

    if (metadata['detector'] == 'PILATUS_6M') and \
       (metadata['size'] == (1679, 1475)):
        metadata['detector'] = 'PILATUS_2M'

    # now compute the filename template and what have you, and also
    # verify that the results stored make sense, particularly w.r.t.
    # the beam centre, which may be stored in pixels not mm.

    template, directory = image2template_directory(image)

    # MAR CCD images record the beam centre in pixels...

    if metadata['detector'] == 'MARCCD':
        metadata['beam'] = (metadata['beam'][0] * metadata['pixel'][0],
                            metadata['beam'][1] * metadata['pixel'][1])

    metadata['directory'] = directory
    metadata['template'] = template
    metadata['start'] = min(matching)
    metadata['end'] = max(matching)
    metadata['matching'] = matching

    return metadata

# FIXME add some unit tests in here...

if __name__ == '__main__':
    import sys
    md = read_image_metadata(sys.argv[1])
    for name in sorted(md):
        print(name, md[name])



#!/usr/bin/env python
# fast_dp.py
#
# A *quick* data reduction jiffy, for the purposes of providing an estimate
# of the quality of a data set as fast as possible along with a set of
# intensities which have been scaled reasonably well. This relies heavily on
# XDS, and forkintegrate in particular.

from __future__ import absolute_import, division, print_function

import copy
import json
import os
import re
import sys
import time
import traceback

import fast_dp
from fast_dp.run_job import get_number_cpus
from fast_dp.cell_spacegroup import check_spacegroup_name, check_split_cell, \
     generate_primitive_cell
import fast_dp.output

from fast_dp.image_readers import read_image_metadata, check_file_readable

from fast_dp.autoindex import autoindex
from fast_dp.integrate import integrate
from fast_dp.scale import scale
from fast_dp.merge import merge
from fast_dp.pointgroup import decide_pointgroup
from fast_dp.logger import write

from optparse import SUPPRESS_HELP, OptionParser

class FastDP:
    '''A class to implement fast data processing for MX beamlines (at Diamond)
    which uses XDS, Pointless, Scala and a couple of other CCP4 odds and
    ends to provide integrated and scaled data in a couple of minutes.'''

    def __init__(self):

        # unguessable input parameters
        self._start_image = None

        # optional user inputs - the unit cell is needed twice, once for the
        # correct lattice and once for the primitive lattice which will
        # be determined from this...

        self._input_spacegroup = None
        self._input_cell = None

        # this is assigned automatically from the input cell above
        self._input_cell_p1 = None

        # user controlled resolution limits
        self._resolution_low = 30.0
        self._resolution_high = 0.0

        # job control see -j -J -k command-line options below for node names
        # see fast_dp#9
        self._n_jobs = 1
        self._n_cores = 0
        self._max_n_jobs = 0
        self._n_cpus = get_number_cpus()
        self._execution_hosts = []

        # image ranges
        self._first_image = None
        self._last_image = None

        # internal data
        self._metadata = None

        # these are the resulting not input ones... clarify this?
        self._p1_unit_cell = None
        self._unit_cell = None
        self._space_group_number = None

        # two additional results

        self._nref = 0
        self._xml_results = None
        self._refined_beam = (0, 0)

    def set_n_jobs(self, n_jobs):
        self._n_jobs = n_jobs

    def set_n_cores(self, n_cores):
        self._n_cores = n_cores

    def set_max_n_jobs(self, max_n_jobs):
        self._max_n_jobs = max_n_jobs

    def set_execution_hosts(self, execution_hosts):
        self._execution_hosts = execution_hosts
        max_n_jobs = 0
        for host in execution_hosts:
            if ':' in host:
                max_n_jobs += int(host.split(':')[1])
        if max_n_jobs:
            self._max_n_jobs = max_n_jobs
        else:
            self._max_n_jobs = len(execution_hosts)
        self._n_jobs = 0

        # add this to the metadata as "extra text"
        et = self._metadata.get('extra_text', '')
        self._metadata['extra_text'] = et + 'CLUSTER_NODES=%s\n' % \
            ' '.join(execution_hosts)

    def get_execution_hosts(self):
        return self._execution_hosts

    def set_first_image(self, first_image):
        self._first_image = first_image

    def set_last_image(self, last_image):
        self._last_image = last_image

    def set_resolution_low(self, resolution_low):
        self._resolution_low = resolution_low

    def set_resolution_high(self, resolution_high):
        self._resolution_high = resolution_high

    def set_start_image(self, start_image):
        '''Set the image to work from: in the majority of cases this will
        be sufficient. This returns a list of image numbers which may be
        missing.'''

        assert(self._start_image is None)

        # check input is image file
        if not os.path.isfile(start_image):
            raise RuntimeError('no image provided: data collection cancelled?')

        check_file_readable(start_image)

        self._start_image = start_image
        self._metadata = read_image_metadata(self._start_image)

        # check that all of the images are present

        matching = self._metadata['matching']

        missing = []

        for j in range(min(matching), max(matching)):
            if not j in matching:
                missing.append(j)

        return missing

    def set_beam(self, beam):
        '''Set the beam centre, in mm, in the Mosflm reference frame.'''

        assert(self._metadata)
        assert(len(beam) == 2)

        self._metadata['beam'] = beam

    def set_distance(self, distance):
        '''Set the detector distance, in mm.'''

        assert(self._metadata)

        self._metadata['distance'] = distance

    def set_atom(self, atom):
        '''Set the heavy atom, if appropriate.'''

        assert(self._metadata)

        self._metadata['atom'] = atom

    # N.B. these two methods assume that the input unit cell etc.
    # has already been tested at the option parsing stage...

    def set_input_spacegroup(self, input_spacegroup):
        self._input_spacegroup = input_spacegroup

    def set_input_cell(self, input_cell):

        self._input_cell = input_cell

        # convert this to a primitive cell based on the centring
        # operation implied by the spacegroup

        assert(self._input_spacegroup)

        self._input_cell_p1 = generate_primitive_cell(
            self._input_cell, self._input_spacegroup).parameters()

    def _read_image_metadata(self):
        '''Get the information from the start image to the metadata bucket.
        For internal use only.'''

        assert(self._start_image)

    def get_metadata_item(self, item):
        '''Get a specific item from the metadata.'''

        assert(self._metadata)
        assert(item in self._metadata)

        return self._metadata[item]

    def process(self):
        '''Main routine, chain together all of the steps imported from
        autoindex, integrate, pointgroup, scale and merge.'''

        try:

            hostname = os.environ['HOSTNAME'].split('.')[0]
            write('Running on: %s' % hostname)

        except Exception:
            pass

        # check input frame limits

        if not self._first_image is None:
            if self._metadata['start'] < self._first_image:
                start = self._metadata['start']
                self._metadata['start'] = self._first_image
                self._metadata['phi_start'] += self._metadata['phi_width'] * \
                                               (self._first_image - start)

        if not self._last_image is None:
            if self._metadata['end'] > self._last_image:
                self._metadata['end'] = self._last_image

        # first if the number of jobs was set to 0, decide something sensible.
        # this should be jobs of a minimum of 5 degrees, 10 frames.

        if self._n_jobs == 0:
            phi = self._metadata['oscillation'][1]

            if phi == 0.0:
                raise RuntimeError('grid scan data')

            wedge = max(10, int(round(5.0 / phi)))
            frames = self._metadata['end'] - self._metadata['start'] + 1
            n_jobs = int(round(frames / wedge))
            if self._max_n_jobs > 0:
                if n_jobs > self._max_n_jobs:
                    n_jobs = self._max_n_jobs
            self.set_n_jobs(n_jobs)

        write('Number of jobs: %d' % self._n_jobs)
        write('Number of cores: %d' % self._n_cores)

        step_time = time.time()

        write('Processing images: %d -> %d' % (self._metadata['start'],
                                               self._metadata['end']))

        phi_end = self._metadata['phi_start'] + self._metadata['phi_width'] * \
                  (self._metadata['end'] - self._metadata['start'] + 1)

        write('Phi range: %.2f -> %.2f' % (self._metadata['phi_start'],
                                           phi_end))

        write('Template: %s' % self._metadata['template'])
        write('Wavelength: %.5f' % self._metadata['wavelength'])
        write('Working in: %s' % os.getcwd())

        try:
            self._p1_unit_cell = autoindex(self._metadata,
                                           input_cell = self._input_cell_p1)
        except Exception as e:
            traceback.print_exc(file = open('fast_dp.error', 'w'))
            write('Autoindexing error: %s' % e)
            return

        try:
            mosaics = integrate(self._metadata, self._p1_unit_cell,
                                self._resolution_low, self._n_jobs,
                                self._n_cores)
            write('Mosaic spread: %.2f < %.2f < %.2f' % tuple(mosaics))
        except RuntimeError as e:
            traceback.print_exc(file = open('fast_dp.error', 'w'))
            write('Integration error: %s' % e)
            return

        try:

            # FIXME in here will need a mechanism to take the input
            # spacegroup, determine the corresponding pointgroup
            # and then apply this (or verify that it is allowed then
            # select)

            metadata = copy.deepcopy(self._metadata)

            cell, sg_num, resol = decide_pointgroup(
                self._p1_unit_cell, metadata,
                input_spacegroup = self._input_spacegroup)
            self._unit_cell = cell
            self._space_group_number = sg_num

            if not self._resolution_high:
                self._resolution_high = resol

        except RuntimeError as e:
            write('Pointgroup error: %s' % e)
            return

        try:
            self._unit_cell, self._space_group, self._nref, beam_pixels = \
            scale(self._unit_cell, self._metadata, self._space_group_number, \
                   self._resolution_high)
            self._refined_beam = (self._metadata['pixel'][1] * beam_pixels[1],
                                  self._metadata['pixel'][0] * beam_pixels[0])

        except RuntimeError as e:
            write('Scaling error: %s' % e)
            return

        try:
            n_images = self._metadata['end'] - self._metadata['start'] + 1
            self._xml_results = merge()
        except RuntimeError as e:
            write('Merging error: %s' % e)
            return

        write('Merging point group: %s' % self._space_group)
        write('Unit cell: %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f' % \
              self._unit_cell)

        duration = time.time() - step_time
        write('Processing took %s (%d s) [%d reflections]' %
              (time.strftime('%Hh %Mm %Ss',
                             time.gmtime(duration)), duration,
               self._nref))
        write('RPS: %.1f' % (float(self._nref) / duration))

        # write out json and xml
        for func in (fast_dp.output.write_json, fast_dp.output.write_ispyb_xml):
          func(self._commandline, self._space_group,
               self._unit_cell, self._xml_results,
               self._start_image, self._refined_beam)

def main():
    '''Main routine for fast_dp.'''

    os.environ['FAST_DP_FORKINTEGRATE'] = '1'

    commandline = ' '.join(sys.argv)

    parser = OptionParser(usage="fast_dp [options] imagefile")

    parser.add_option("-?", action="help", help=SUPPRESS_HELP)

    parser.add_option('-b', '--beam', dest = 'beam',
                      help = 'Beam centre: x, y (mm)')

    parser.add_option('-d', '--distance', dest = 'distance',
                      help = 'Detector distance: d (mm)')

    parser.add_option('-a', '--atom', dest = 'atom',
                      help = 'Atom type (e.g. Se)')

    parser.add_option('-j', '--number-of-jobs', dest = 'number_of_jobs',
                      help = 'Number of jobs for integration')
    parser.add_option('-k', '--number-of-cores', dest = 'number_of_cores',
                      help = 'Number of cores for integration')
    parser.add_option('-J', '--maximum-number-of-jobs',
                      dest = 'maximum_number_of_jobs',
                      help = 'Maximum number of jobs for integration')
    parser.add_option('-e', '--execution-hosts',
                      dest = 'execution_hosts',
                      help = 'names for execution hosts for forkxds')

    parser.add_option('-c', '--cell', dest = 'cell',
                      help = 'Cell constants for processing, needs spacegroup')
    parser.add_option('-s', '--spacegroup', dest = 'spacegroup',
                      help = 'Spacegroup for scaling and merging')

    parser.add_option('-1', '--first-image', dest = 'first_image',
                      help = 'First image for processing')
    parser.add_option('-N', '--last-image', dest = 'last_image',
                      help = 'First image for processing')

    parser.add_option('-r', '--resolution-high', dest = 'resolution_high',
                      help = 'High resolution limit')
    parser.add_option('-R', '--resolution-low', dest = 'resolution_low',
                      help = 'Low resolution limit')

    parser.add_option('-l', '--lib-name', dest = 'lib_name',
                      help = 'HDF5 reader library (i.e. neggia etc.)')

    parser.add_option('--version', dest='version', action='store_true',
                      default=False, help='Print fast_dp version')

    (options, args) = parser.parse_args()

    if options.version:
      print('Fast_DP version %s' % fast_dp.__version__)
      sys.exit(0)

    if len(args) != 1:
      parser.error("You must point to one image of the dataset to process")

    image = args[0]

    xia2_format = re.match(r"^(.*):(\d+):(\d+)$", image)
    if xia2_format:
      # Image can be given in xia2-style format, ie.
      #   set_of_images_00001.cbf:1:5000
      # to select images 1 to 5000. Resolve any conflicts
      # with -1/-N in favour of the explicit arguments.
      image = xia2_format.group(1)
      if not options.first_image:
        options.first_image = xia2_format.group(2)
      if not options.last_image:
        options.last_image = xia2_format.group(3)

    if options.lib_name:
        from image_readers import set_lib_name
        set_lib_name(options.lib_name)

    try:
        write('Fast_DP version %s' % fast_dp.__version__)
        finst = FastDP()
        finst._commandline = commandline
        write('Starting image: %s' % image)
        missing = finst.set_start_image(image)
        if options.beam:
            x, y = tuple(map(float, options.beam.split(',')))
            finst.set_beam((x, y))

        if options.distance:
            finst.set_distance(float(options.distance))

        if options.atom:
            finst.set_atom(options.atom)

        if options.maximum_number_of_jobs:
            finst.set_max_n_jobs(int(options.maximum_number_of_jobs))

        if options.execution_hosts:
            finst.set_execution_hosts(options.execution_hosts.split(','))
            write('Execution hosts: %s' % ' '.join(finst.get_execution_hosts()))

        if options.number_of_jobs:
            if options.maximum_number_of_jobs:
                n_jobs = min(int(options.number_of_jobs),
                             int(options.maximum_number_of_jobs))
            else:
                n_jobs = int(options.number_of_jobs)
            finst.set_n_jobs(n_jobs)

        if options.number_of_cores:
            n_cores = int(options.number_of_cores)
            finst.set_n_cores(n_cores)

        if options.first_image:
            first_image = int(options.first_image)
            missing = [m for m in missing if m >= first_image]
            finst.set_first_image(first_image)

        if options.last_image:
            last_image = int(options.last_image)
            missing = [m for m in missing if m <= last_image]
            finst.set_last_image(last_image)

        if missing:
            raise RuntimeError('images missing: %s' % \
                ' '.join(map(str, missing)))

        if options.resolution_low:
            finst.set_resolution_low(float(options.resolution_low))

        if options.resolution_high:
            finst.set_resolution_high(float(options.resolution_high))

        # must input spacegroup first as unpacking of the unit cell
        # will depend on the centering operation...

        if options.spacegroup:
            try:
                spacegroup = check_spacegroup_name(options.spacegroup)
                finst.set_input_spacegroup(spacegroup)
                write('Set spacegroup: %s' % spacegroup)
            except RuntimeError:
                write('Spacegroup %s not recognised: ignoring' % \
                      options.spacegroup)

        if options.cell:
            assert(options.spacegroup)
            cell = check_split_cell(options.cell)
            write('Set cell: %.2f %.2f %.2f %.2f %.2f %.2f' % cell)
            finst.set_input_cell(cell)

        finst.process()

    except Exception as e:
        traceback.print_exc(file = open('fast_dp.error', 'w'))
        write('Fast DP error: %s' % str(e))

    json_stuff = { }
    for prop in dir(finst):
        ignore = ['_read_image_metadata']
        if not prop.startswith('_') or prop.startswith('__'):
            continue
        if prop in ignore:
            continue
        json_stuff[prop] = getattr(finst, prop)
    with open('fast_dp.state', 'wb') as fh:
      json.dump(json_stuff, fh)

if __name__ == '__main__':
    main()



# header2edna_xml - a jiffy to read a Diffraction Image from an instrument
# here at Diamond Light Source and generate EDNA xml.

from __future__ import absolute_import, division, print_function

import os
import sys

from image_readers import read_image_metadata

from image_names import image2image

# need to be sure that I have the following:
#
# binning e.g. 2x2
# image saturation i.e. 65535
# detector name i.e. ADSC Q315 bin 2x2
# image dimensions in pixels
# pixel sizes in mm
# oscillation start and width
# exposure time
# distance
# wavelength
# beam position in mm in the mosflm reference frame


# FIXME add a hash table here for the detector long names and short names...

detector_short_names = {
    ('PILATUS_2M', 1679):'p2m',
    ('PILATUS_6M', 2527):'p6m',
    ('ADSC', 3072):'q315-2x',
    ('RIGAKU', 3000):'raxis4'
    }

detector_long_names = {
    ('PILATUS_2M', 1679):'PILATUS 2M',
    ('PILATUS_6M', 2527):'PILATUS 6M',
    ('ADSC', 3072):'ADSC Q315 bin 2x2',
    ('RIGAKU', 3000):'RIGAKU RAXIS IV'
    }

def header2edna_xml(image_file, minosc, mintime):
    '''Read an image header, generate EDNA xml. DLS #1295.'''

    template = os.path.join(os.environ['FAST_DP_ROOT'],
                            'lib', 'templates',
                            'EDNA_HEADER_XML.INP')

    if not os.path.exists(template):
        raise RuntimeError('template for EDNA_HEADER_XML.INP cannot be found')

    edna_xml_for_subwedge = open(template, 'r').read().strip()

    header = read_image_metadata(image_file)

    long_name = detector_long_names[(header['detector'], header['size'][0])]
    short_name = detector_short_names[(header['detector'], header['size'][0])]

    xml = edna_xml_for_subwedge.format(
        exposure_time = header['exposure_time'],
        minexptime = mintime,
        wavelength = header['wavelength'],
        beam_x = header['beam'][0],
        beam_y = header['beam'][1],
        bin = header.get('bin', '1x1'),
        distance = header['distance'],
        saturation = header.get('saturation', 65535),
        detector_name = long_name,
        pixel_x = header['size'][0],
        pixel_y = header['size'][1],
        pixel_size_x = header['pixel'][0],
        pixel_size_y = header['pixel'][1],
        detector_type = short_name,
        maxoscspeed = 10.0,
        minoscwidth = minosc,
        osc_width = header['phi_width'],
        oscaxis = 'phi',
        osc_end = header['phi_end'],
        osc_start = header['phi_start'],
        image_number = image2image(image_file),
        image_name = image_file)

    return xml

if __name__ == '__main__':

    if len(sys.argv) == 4:
        minosc = float(sys.argv[2])
        mintime = float(sys.argv[3])
    else:
        minosc = 0.1
        mintime = 0.067

    xml = header2edna_xml(sys.argv[1], minosc, mintime)

    print(xml)
    # .replace('><', '>\n<')



#!/usr/bin/env python
# fast_rdp.py
#
# Re-data-process i.e. repeat the finishing stages of a fast_dp job, to adjust
# the resolution limit or assign some other symmetry.

from __future__ import absolute_import, division, print_function

import json
import sys
import os
import time
import copy
import traceback

from run_job import get_number_cpus
from cell_spacegroup import check_spacegroup_name, check_split_cell, \
     generate_primitive_cell
import output

from image_readers import read_image_metadata, check_file_readable

from autoindex import autoindex
from integrate import integrate
from scale import scale
from merge import merge
from pointgroup import decide_pointgroup
from logger import write, set_filename
set_filename('fast_rdp.log')

class FastRDP:
    '''A class to implement fast data processing for MX beamlines (at Diamond)
    which uses XDS, Pointless, Scala and a couple of other CCP4 odds and
    ends to provide integrated and scaled data in a couple of minutes.'''

    def __init__(self):

        with open('fast_dp.state', 'rb') as fh:
          json_stuff = json.load(fh)
        for prop in json_stuff:
            # do not want to pass this along since that will limit what we
            # can reindex to...
            if prop == '_input_spacegroup':
                self._input_spacegroup = None
                continue
            setattr(self, prop, json_stuff[prop])

    def set_first_image(self, first_image):
        self._first_image = first_image

    def set_last_image(self, last_image):
        self._last_image = last_image

    def set_resolution_low(self, resolution_low):
        self._resolution_low = resolution_low

    def set_resolution_high(self, resolution_high):
        self._resolution_high = resolution_high

    def set_atom(self, atom):
        '''Set the heavy atom, if appropriate.'''

        assert(self._metadata)

        self._metadata['atom'] = atom

    # N.B. these two methods assume that the input unit cell etc.
    # has already been tested at the option parsing stage...

    def set_input_spacegroup(self, input_spacegroup):
        self._input_spacegroup = input_spacegroup

    def set_input_cell(self, input_cell):

        self._input_cell = input_cell

        # convert this to a primitive cell based on the centring
        # operation implied by the spacegroup

        assert(self._input_spacegroup)

        self._input_cell_p1 = generate_primitive_cell(
            self._input_cell, self._input_spacegroup).parameters()

        # FIXME for reprocessing purposes verify here that the input unit cell
        # matches the one which was used for previous fast_dp job - check
        # self._p1_unit_cell

    def get_metadata_item(self, item):
        '''Get a specific item from the metadata.'''

        assert(self._metadata)
        assert(item in self._metadata)

        return self._metadata[item]

    def reprocess(self):
        '''Main routine, chain together last few steps of processing i.e.
        pointgroup, scale and merge.'''

        try:

            hostname = os.environ['HOSTNAME'].split('.')[0]
            write('Running on: %s' % hostname)

        except Exception:
            pass

        # check input frame limits

        if not self._first_image is None:
            if self._metadata['start'] < self._first_image:
                start = self._metadata['start']
                self._metadata['start'] = self._first_image
                self._metadata['phi_start'] += self._metadata['phi_width'] * \
                                               (self._first_image - start)

        if not self._last_image is None:
            if self._metadata['end'] > self._last_image:
                self._metadata['end'] = self._last_image

        step_time = time.time()

        write('Processing images: %d -> %d' % (self._metadata['start'],
                                               self._metadata['end']))

        phi_end = self._metadata['phi_start'] + self._metadata['phi_width'] * \
                  (self._metadata['end'] - self._metadata['start'] + 1)

        write('Phi range: %.2f -> %.2f' % (self._metadata['phi_start'],
                                           phi_end))

        write('Template: %s' % self._metadata['template'])
        write('Wavelength: %.5f' % self._metadata['wavelength'])
        write('Working in: %s' % os.getcwd())

        # just for information for the user, print all options for indexing
        # FIXME should be able to run the same from CORRECT.LP which would
        # work better....

        from xds_reader import read_xds_idxref_lp
        from cell_spacegroup import spacegroup_to_lattice

        results = read_xds_idxref_lp('IDXREF.LP')

        write('For reference, all indexing results:')
        write('%3s %6s %6s %6s %6s %6s %6s' % \
            ('Lattice', 'a', 'b', 'c', 'alpha', 'beta', 'gamma'))

        for r in reversed(sorted(results)):
            if not type(r) == type(1):
                continue
            cell = results[r][1]
            write('%7s %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f' % \
                (spacegroup_to_lattice(r), cell[0], cell[1], cell[2],
                cell[3], cell[4], cell[5]))

        try:

            # FIXME in here will need a mechanism to take the input
            # spacegroup, determine the corresponding pointgroup
            # and then apply this (or verify that it is allowed then
            # select)

            metadata = copy.deepcopy(self._metadata)

            cell, sg_num, resol = decide_pointgroup(
                self._p1_unit_cell, metadata,
                input_spacegroup = self._input_spacegroup)
            self._unit_cell = cell
            self._space_group_number = sg_num

            if not self._resolution_high:
                self._resolution_high = resol

        except RuntimeError as e:
            write('Pointgroup error: %s' % e)
            return

        try:
            self._unit_cell, self._space_group, self._nref, beam_pixels = \
            scale(self._unit_cell, self._metadata, self._space_group_number, \
                   self._resolution_high)
            self._refined_beam = (self._metadata['pixel'][1] * beam_pixels[1],
                                  self._metadata['pixel'][0] * beam_pixels[0])

        except RuntimeError as e:
            write('Scaling error: %s' % e)
            return

        try:
            n_images = self._metadata['end'] - self._metadata['start'] + 1
            self._xml_results = merge(hklout='fast_rdp.mtz',
                                      aimless_log='aimless_rerun.log')
        except RuntimeError as e:
            write('Merging error: %s' % e)
            return

        write('Merging point group: %s' % self._space_group)
        write('Unit cell: %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f' % \
              self._unit_cell)

        duration = time.time() - step_time
        write('Reprocessing took %s (%d s) [%d reflections]' %
              (time.strftime('%Hh %Mm %Ss',
                             time.gmtime(duration)), duration,
               self._nref))

        # write out json and xml
        for func, filename in [ (output.write_json, 'fast_rdp.json'),
                                (output.write_ispyb_xml, 'fast_rdp.xml') ]:
          func(self._commandline, self._space_group,
               self._unit_cell, self._xml_results,
               self._start_image, self._refined_beam,
               filename=filename)

def main():
    '''Main routine for fast_rdp.'''

    os.environ['FAST_DP_FORKINTEGRATE'] = '1'

    from optparse import OptionParser

    commandline = ' '.join(sys.argv)

    parser = OptionParser()

    parser.add_option('-a', '--atom', dest = 'atom',
                      help = 'Atom type (e.g. Se)')

    parser.add_option('-c', '--cell', dest = 'cell',
                      help = 'Cell constants for processing, needs spacegroup')
    parser.add_option('-s', '--spacegroup', dest = 'spacegroup',
                      help = 'Spacegroup for scaling and merging')

    parser.add_option('-1', '--first-image', dest = 'first_image',
                      help = 'First image for processing')
    parser.add_option('-N', '--last-image', dest = 'last_image',
                      help = 'First image for processing')

    parser.add_option('-r', '--resolution-high', dest = 'resolution_high',
                      help = 'High resolution limit')
    parser.add_option('-R', '--resolution-low', dest = 'resolution_low',
                      help = 'Low resolution limit')

    (options, args) = parser.parse_args()

    assert(len(args) < 2)

    # if arg given then assume that this is a directory with a fast_dp
    # job it in, but where $user does not have access to write - so first
    # copy the files needed across

    if len(args) == 1:
        if not os.path.isdir(args[0]):
            raise RuntimeError('in this mode, provide /path/to/fast_dp/dir')
        from_dir = args[0]
        for filename in os.listdir(from_dir):
            if os.path.isdir(os.path.join(from_dir, filename)):
                continue
            import shutil
            shutil.copyfile(os.path.join(from_dir, filename),
                            os.path.join(os.getcwd(), filename))
    else:
        from_dir = None

    try:
        fast_rdp = FastRDP()
        fast_rdp._commandline = commandline
        write('Fast_RDP installed in: %s' % os.environ['FAST_DP_ROOT'])
        write('Working in: %s' % os.getcwd())
        if from_dir:
            write('Working from: %s' % from_dir)

        if options.atom:
            fast_rdp.set_atom(options.atom)

        if options.first_image:
            first_image = int(options.first_image)
            fast_rdp.set_first_image(first_image)

        if options.last_image:
            last_image = int(options.last_image)
            fast_rdp.set_last_image(last_image)

        if options.resolution_low:
            fast_rdp.set_resolution_low(float(options.resolution_low))

        if options.resolution_high:
            fast_rdp.set_resolution_high(float(options.resolution_high))

        # must input spacegroup first as unpacking of the unit cell
        # will depend on the centering operation...

        if options.spacegroup:
            try:
                spacegroup = check_spacegroup_name(options.spacegroup)
                fast_rdp.set_input_spacegroup(spacegroup)
                write('Set spacegroup: %s' % spacegroup)
            except RuntimeError as e:
                write('Spacegroup %s not recognised: ignoring' % \
                      options.spacegroup)

        if options.cell:
            assert(options.spacegroup)
            cell = check_split_cell(options.cell)
            write('Set cell: %.2f %.2f %.2f %.2f %.2f %.2f' % cell)
            fast_rdp.set_input_cell(cell)

        fast_rdp.reprocess()

    except Exception as e:
        traceback.print_exc(file = open('fast_rdp.error', 'w'))
        write('Fast RDP error: %s' % str(e))

if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function

import os
import pkg_resources

from fast_dp.run_job import get_number_cpus

# XDS.INP writer functions - two (three) of these, to write out commands
# for autoindexing, integration then postrefinement and scaling. Split
# up thus because XDS will frequently stop after autoindexing complaining
# that your data are not perfect, and then you probably want to run post-
# refinement and scaling a couple of times. With the latter need to be able
# to control the scale factors applied. N.B. these calculate the image
# ranges to use from the input metadata.

def get_template(detector, instruction):
  '''Read the template for a given detector from the package resource files.'''
  template = 'templates/%s_%s.INP' % (detector, instruction)
  if not pkg_resources.resource_exists('fast_dp', template):
    raise RuntimeError('{instruction} template for {detector} not found'.format(
        instruction=instruction, detector=detector))

  return pkg_resources.resource_string('fast_dp', template).decode('utf-8').strip()


def write_xds_inp_autoindex(metadata, xds_inp):
  template_str = get_template(metadata['detector'], 'INDEX')
  with open(xds_inp, 'w') as fout:

    # should somehow hang this from an anomalous flag

    friedels_law = 'FALSE'

    fout.write('%s\n' % template_str.format(
        extra_text = metadata.get('extra_text', '!PARAMETER=VALUE'),
        no_processors = get_number_cpus(),
        nx = metadata['size'][0],
        ny = metadata['size'][1],
        qx = metadata['pixel'][0],
        qy = metadata['pixel'][1],
        orgx = metadata['beam'][0] / metadata['pixel'][0],
        orgy = metadata['beam'][1] / metadata['pixel'][1],
        distance = metadata['distance'],
        sensor = metadata.get('sensor', None),
        wavelength = metadata['wavelength'],
        oscillation = metadata['oscillation'][1],
        friedels_law = friedels_law,
        template = os.path.join(metadata['directory'],
                                metadata['template'].replace('#', '?')),
        starting_angle = metadata['oscillation'][0],
        starting_image = metadata['start']))

    # then we get the non-template stuff

    fout.write('DATA_RANGE=%d %d\n' % (metadata['start'],
                                       metadata['end']))

    # compute the background range as min(all, 5) #TODO maybe 5 degrees?

    if metadata['end'] - metadata['start'] > 5:
        fout.write('BACKGROUND_RANGE=%d %d\n' % \
                   (metadata['start'], metadata['start'] + 5))
    else:
        fout.write('BACKGROUND_RANGE=%d %d\n' % (metadata['start'],
                                                 metadata['end']))

    # REFINE(IDXREF)=
    fout.write('REFINE(IDXREF)=CELL AXIS ORIENTATION POSITION BEAM\n')

    # by default autoindex off all images - can make this better later on.
    # Ok: I think it is too slow already. Three wedges, as per xia2...
    # that would be 5 images per wedge, then. Erk. Should be *degrees*

    images = range(metadata['start'], metadata['end'] + 1)

    wedge_size = int(round(5.0  / metadata['oscillation'][1])) - 1

    wedge = (images[0], images[0] + wedge_size)
    fout.write('SPOT_RANGE=%d %d\n' % wedge)

    # if we have more than 90 degrees of data, use wedges at the start,
    # 45 degrees in and 90 degrees in, else use a wedge at the start,
    # one in the middle and one at the end.

    # if less than 15 degrees of data, use all of the images

    if (metadata['end'] - metadata['start']) * metadata['oscillation'][1] < 15:
        fout.write('SPOT_RANGE=%d %d\n' % (metadata['start'],
                                           metadata['end']))

    elif int(90.0 / metadata['oscillation'][1]) + wedge_size in images:
        wedge = (int(45.0 / metadata['oscillation'][1]),
                 int(45.0 / metadata['oscillation'][1]) + wedge_size)
        fout.write('SPOT_RANGE=%d %d\n' % wedge)
        wedge = (int(90.0 / metadata['oscillation'][1]),
                 int(90.0 / metadata['oscillation'][1]) + wedge_size)
        fout.write('SPOT_RANGE=%d %d\n' % wedge)

    else:
        mid = (len(images) / 2) - wedge_size + images[0] - 1
        wedge = (mid, mid + wedge_size)
        fout.write('SPOT_RANGE=%d %d\n' % wedge)
        wedge = (images[-wedge_size], images[-1])
        fout.write('SPOT_RANGE=%d %d\n' % wedge)


def write_xds_inp_autoindex_p1_cell(metadata, xds_inp, cell):
  template_str = get_template(metadata['detector'], 'INDEX')
  with open(xds_inp, 'w') as fout:
    # should somehow hang this from an anomalous flag

    friedels_law = 'FALSE'

    fout.write('%s\n' % template_str.format(
        extra_text = metadata.get('extra_text', '!PARAMETER=VALUE'),
        no_processors = get_number_cpus(),
        nx = metadata['size'][0],
        ny = metadata['size'][1],
        qx = metadata['pixel'][0],
        qy = metadata['pixel'][1],
        orgx = metadata['beam'][0] / metadata['pixel'][0],
        orgy = metadata['beam'][1] / metadata['pixel'][1],
        distance = metadata['distance'],
        sensor = metadata.get('sensor', None),
        wavelength = metadata['wavelength'],
        oscillation = metadata['oscillation'][1],
        friedels_law = friedels_law,
        template = os.path.join(metadata['directory'],
                                metadata['template'].replace('#', '?')),
        starting_angle = metadata['oscillation'][0],
        starting_image = metadata['start']))

    # cell, spacegroup

    fout.write('SPACE_GROUP_NUMBER=1\n')
    fout.write('UNIT_CELL_CONSTANTS=%f %f %f %f %f %f\n' % tuple(cell))

    # then we get the non-template stuff

    fout.write('DATA_RANGE=%d %d\n' % (metadata['start'],
                                       metadata['end']))

    # compute the background range as min(all, 5)

    if metadata['end'] - metadata['start'] > 5:
        fout.write('BACKGROUND_RANGE=%d %d\n' % \
                   (metadata['start'], metadata['start'] + 5))
    else:
        fout.write('BACKGROUND_RANGE=%d %d\n' % (metadata['start'],
                                                 metadata['end']))

    # by default autoindex off all images - can make this better later on.
    # Ok: I think it is too slow already. Three wedges, as per xia2...
    # that would be 5 images per wedge, then. Erk. Should be *degrees*

    images = range(metadata['start'], metadata['end'] + 1)

    wedge_size = int(round(5.0  / metadata['oscillation'][1])) - 1

    wedge = (images[0], images[0] + wedge_size)
    fout.write('SPOT_RANGE=%d %d\n' % wedge)

    # if we have more than 90 degrees of data, use wedges at the start,
    # 45 degrees in and 90 degrees in, else use a wedge at the start,
    # one in the middle and one at the end.

    # if less than 15 degrees of data, use all of the images

    if (metadata['end'] - metadata['start']) * metadata['oscillation'][1] < 15:
        fout.write('SPOT_RANGE=%d %d\n' % (metadata['start'],
                                           metadata['end']))

    elif int(90.0 / metadata['oscillation'][1]) + wedge_size in images:
        wedge = (int(45.0 / metadata['oscillation'][1]),
                 int(45.0 / metadata['oscillation'][1]) + wedge_size)
        fout.write('SPOT_RANGE=%d %d\n' % wedge)
        wedge = (int(90.0 / metadata['oscillation'][1]),
                 int(90.0 / metadata['oscillation'][1]) + wedge_size)
        fout.write('SPOT_RANGE=%d %d\n' % wedge)

    else:
        mid = (len(images) / 2) - wedge_size + images[0] - 1
        wedge = (mid, mid + wedge_size)
        fout.write('SPOT_RANGE=%d %d\n' % wedge)
        wedge = (images[-5], images[-1])
        fout.write('SPOT_RANGE=%d %d\n' % wedge)


def write_xds_inp_integrate(metadata, xds_inp, resolution_low, no_jobs=1, no_processors=0):
  template_str = get_template(metadata['detector'], 'INTEGRATE')
  with open(xds_inp, 'w') as fout:

    # FIXME in here calculate the maximum number of jobs to correspond at the
    # least to 5 degree wedges / job.

    # should somehow hang this from an anomalous flag

    friedels_law = 'FALSE'

    if no_processors == 0:
        no_processors = get_number_cpus()

    fout.write('%s\n' % template_str.format(
        extra_text = metadata.get('extra_text', '!PARAMETER=VALUE'),
        no_processors = no_processors,
        no_jobs = no_jobs,
        resolution_low = resolution_low,
        resolution_high = 0.0,
        nx = metadata['size'][0],
        ny = metadata['size'][1],
        qx = metadata['pixel'][0],
        qy = metadata['pixel'][1],
        orgx = metadata['beam'][0] / metadata['pixel'][0],
        orgy = metadata['beam'][1] / metadata['pixel'][1],
        distance = metadata['distance'],
        sensor = metadata.get('sensor', None),
        wavelength = metadata['wavelength'],
        oscillation = metadata['oscillation'][1],
        friedels_law = friedels_law,
        template = os.path.join(metadata['directory'],
                                metadata['template'].replace('#', '?')),
        starting_angle = metadata['oscillation'][0],
        starting_image = metadata['start']))

    # then we get the non-template stuff

    fout.write('DATA_RANGE=%d %d\n' % (metadata['start'],
                                       metadata['end']))


# N.B. this one is a little different to the others as the inclusion of
# the cell constants and symmetry are *mandatory*. N.B. default may be
# to use the triclinic solution in the first pass.

def write_xds_inp_correct(metadata, unit_cell, space_group_number,
                          xds_inp, scale = True,
                          resolution_low = 30, resolution_high = 0.0,
                          turn_subset = False):
  template_str = get_template(metadata['detector'], 'CORRECT')
  with open(xds_inp, 'w') as fout:
    # should somehow hang this from an anomalous flag

    if 'atom' in metadata:
        friedels_law = 'FALSE'
    else:
        friedels_law = 'TRUE'

    if scale:
        corrections = 'ALL'
    else:
        corrections = '!'

    fout.write('%s\n' % template_str.format(
        extra_text = metadata.get('extra_text', '!PARAMETER=VALUE'),
        no_processors = get_number_cpus(),
        resolution_low = resolution_low,
        resolution_high = resolution_high,
        unit_cell_a = unit_cell[0],
        unit_cell_b = unit_cell[1],
        unit_cell_c = unit_cell[2],
        unit_cell_alpha = unit_cell[3],
        unit_cell_beta = unit_cell[4],
        unit_cell_gamma = unit_cell[5],
        space_group_number = space_group_number,
        nx = metadata['size'][0],
        ny = metadata['size'][1],
        qx = metadata['pixel'][0],
        qy = metadata['pixel'][1],
        orgx = metadata['beam'][0] / metadata['pixel'][0],
        orgy = metadata['beam'][1] / metadata['pixel'][1],
        distance = metadata['distance'],
        sensor = metadata.get('sensor', None),
        wavelength = metadata['wavelength'],
        oscillation = metadata['oscillation'][1],
        friedels_law = friedels_law,
        corrections = corrections,
        template = os.path.join(metadata['directory'],
                                metadata['template'].replace('#', '?')),
        starting_angle = metadata['oscillation'][0],
        starting_image = metadata['start']))

    # then we get the non-template stuff

    if turn_subset:
        # limit to 360 degrees...
        width = metadata['oscillation'][1]
        start, end = metadata['start'], metadata['end']
        if (end - start + 1) * width > 360:
            end = start + (360. / width) - 1
        fout.write('DATA_RANGE=%d %d\n' % (start, end))
    else:
        fout.write('DATA_RANGE=%d %d\n' % (metadata['start'], metadata['end']))


def write_xds_inp_correct_no_cell(metadata,
                                  xds_inp, scale = True,
                                  resolution_low = 30, resolution_high = 0.0):
  template_str = get_template(metadata['detector'], 'CORRECT_NO_CELL')
  with open(xds_inp, 'w') as fout:
    # should somehow hang this from an anomalous flag

    friedels_law = 'FALSE'

    if scale:
        corrections = 'ALL'
    else:
        corrections = '!'

    fout.write('%s\n' % template_str.format(
        extra_text = metadata.get('extra_text', '!PARAMETER=VALUE'),
        no_processors = get_number_cpus(),
        resolution_low = resolution_low,
        resolution_high = resolution_high,
        nx = metadata['size'][0],
        ny = metadata['size'][1],
        qx = metadata['pixel'][0],
        qy = metadata['pixel'][1],
        orgx = metadata['beam'][0] / metadata['pixel'][0],
        orgy = metadata['beam'][1] / metadata['pixel'][1],
        distance = metadata['distance'],
        sensor = metadata['sensor'],
        wavelength = metadata['wavelength'],
        oscillation = metadata['oscillation'][1],
        friedels_law = friedels_law,
        corrections = corrections,
        template = os.path.join(metadata['directory'],
                                metadata['template'].replace('#', '?')),
        starting_angle = metadata['oscillation'][0],
        starting_image = metadata['start']))

    # then we get the non-template stuff

    fout.write('DATA_RANGE=%d %d\n' % (metadata['start'],
                                       metadata['end']))



from __future__ import absolute_import, division, print_function

import os

class _writer:
    '''A specialist class to write to the screen and fast_dp.log.'''

    def __init__(self):
        self._fout = None
        self._filename = 'fast_dp.log'
        return

    def set_filename(self, filename):
        self._filename = filename

    def __del__(self):
        if self._fout:
            self._fout.close()
        self._fout = None
        return

    def __call__(self, record):
        self.write(record)

    def write(self, record):
        if not self._fout:
            self._fout = open(self._filename, 'w')

        self._fout.write('%s\n' % record)
        print(record)
        return

write = _writer()

def set_filename(filename):
    write.set_filename(filename)



from __future__ import absolute_import, division, print_function

import fast_dp.output

def test_ispyb_xml_template_is_accessible():
  xml = fast_dp.output.get_ispyb_template()
  assert '<AutoProcContainer>' in xml



from __future__ import absolute_import, division, print_function

import os
import shutil

from fast_dp.xds_writer import write_xds_inp_autoindex, write_xds_inp_autoindex_p1_cell
from fast_dp.xds_reader import read_xds_idxref_lp
from fast_dp.run_job import run_job

from fast_dp.cell_spacegroup import spacegroup_to_lattice

from fast_dp.logger import write

def autoindex(metadata, input_cell = None):
    '''Perform the autoindexing, using metatdata, get a list of possible
    lattices and record / return the triclinic cell constants (get these from
    XPARM.XDS).'''

    assert(metadata)

    xds_inp = 'AUTOINDEX.INP'

    if input_cell:
        write_xds_inp_autoindex_p1_cell(metadata, xds_inp, input_cell)
    else:
        write_xds_inp_autoindex(metadata, xds_inp)

    shutil.copyfile(xds_inp, 'XDS.INP')

    log = run_job('xds_par')

    with open('autoindex.log', 'w') as fout:
        fout.write(''.join(log))

    # sequentially check for errors... XYCORR INIT COLSPOT IDXREF

    for step in ['XYCORR', 'INIT', 'COLSPOT', 'IDXREF']:
        lastrecord = open('%s.LP' % step).readlines()[-1]
        if '!!! ERROR !!!' in lastrecord:
            raise RuntimeError('error in %s: %s' % \
                  (step, lastrecord.replace('!!! ERROR !!!', '').strip()))

    results = read_xds_idxref_lp('IDXREF.LP')

    # FIXME if input cell was given, verify that this is an allowed
    # permutation. If it was not, raise a RuntimeError. This remains to be
    # fixed

    write('All autoindexing results:')
    write('%3s %6s %6s %6s %6s %6s %6s' % \
          ('Lattice', 'a', 'b', 'c', 'alpha', 'beta', 'gamma'))

    for r in reversed(sorted(results)):
        if not type(r) == type(1):
            continue
        cell = results[r][1]
        write('%7s %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f' % \
              (spacegroup_to_lattice(r), cell[0], cell[1], cell[2],
               cell[3], cell[4], cell[5]))

    # should probably print this for debuging

    try:
        return results[1][1]
    except:
        raise RuntimeError('getting P1 cell for autoindex')



from __future__ import absolute_import, division, print_function

import xml.dom.minidom

from fast_dp.cell_spacegroup import lauegroup_to_lattice
from cctbx import sgtbx

def read_pointless_xml(pointless_xml_file):
    '''Read through the pointless xml output, return as a list of spacegroup
    numbers in order of likelihood, corresponding to the pointgroup of the
    data.'''

    dom = xml.dom.minidom.parse(pointless_xml_file)

    best = dom.getElementsByTagName('BestSolution')[0]
    pointgroup = best.getElementsByTagName(
        'GroupName')[0].childNodes[0].data
    confidence = float(best.getElementsByTagName(
        'Confidence')[0].childNodes[0].data)
    totalprob = float(best.getElementsByTagName(
        'TotalProb')[0].childNodes[0].data)
    reindex_matrix = map(float, best.getElementsByTagName(
        'ReindexMatrix')[0].childNodes[0].data.split())
    reindex_operator = best.getElementsByTagName(
        'ReindexOperator')[0].childNodes[0].data.strip()

    scorelist = dom.getElementsByTagName('LaueGroupScoreList')[0]
    scores = scorelist.getElementsByTagName('LaueGroupScore')

    results = []

    for s in scores:
        number = int(s.getElementsByTagName(
            'number')[0].childNodes[0].data)
        lauegroup = str(s.getElementsByTagName(
            'LaueGroupName')[0].childNodes[0].data).strip()
        if lauegroup[0] == 'H':
            lauegroup = 'R%s' % lauegroup[1:]
        pointgroup = sgtbx.space_group_type(lauegroup).group(
            ).build_derived_acentric_group().type().number()
        reindex = s.getElementsByTagName(
            'ReindexOperator')[0].childNodes[0].data
        netzc = float(s.getElementsByTagName(
            'NetZCC')[0].childNodes[0].data)
        likelihood = float(s.getElementsByTagName(
            'Likelihood')[0].childNodes[0].data)
        r_merge = float(s.getElementsByTagName(
            'R')[0].childNodes[0].data)
        delta = float(s.getElementsByTagName(
            'CellDelta')[0].childNodes[0].data)

        # record this as a possible lattice... if it's Z score
        # is positive, anyway - except this does kinda bias towards
        # those lattices which have all symops => no Z-.

        lattice = lauegroup_to_lattice(lauegroup)
        if netzc >= 0.0:
            results.append((lattice, pointgroup))

    return results



from __future__ import absolute_import, division, print_function

import pkg_resources
import json
import os

def write_json(commandline, spacegroup, unit_cell,
               xml_results, start_image, refined_beam,
               filename='fast_dp.json'):
  '''Write out nice JSON for downstream processing.'''

  with open(filename, 'w') as fh:
    json.dump({ 'commandline': commandline,
                'refined_beam': refined_beam,
                'spacegroup': spacegroup,
                'unit_cell': unit_cell,
                'xml_results': xml_results,
              }, fh, sort_keys=True,
                 indent=2, separators=(',', ': '))

def get_ispyb_template():
  '''Read the ispyb.xml template from the package resources.'''
  xml_template = pkg_resources.resource_string('fast_dp', 'templates/ispyb.xml').decode('utf-8')
  assert xml_template, 'Error retrieving XML template'
  return xml_template

def write_ispyb_xml(commandline, spacegroup, unit_cell,
                    xml_results, start_image, refined_beam,
                    filename='fast_dp.xml'):
  '''Write out big lump of XML for ISPyB import.'''

  xml_template = get_ispyb_template()

  with open(filename, 'w') as fh:
    fh.write(xml_template.format(
        commandline = commandline,
        spacegroup = spacegroup,
        cell_a = unit_cell[0],
        cell_b = unit_cell[1],
        cell_c = unit_cell[2],
        cell_alpha = unit_cell[3],
        cell_beta = unit_cell[4],
        cell_gamma = unit_cell[5],
        results_directory = os.getcwd(),
        rmerge_overall = xml_results['rmerge_overall'],
        rmeas_overall = xml_results['rmeas_overall'],
        resol_high_overall = xml_results['resol_high_overall'],
        resol_low_overall = xml_results['resol_low_overall'],
        isigma_overall = xml_results['isigma_overall'],
        completeness_overall = xml_results['completeness_overall'],
        multiplicity_overall = xml_results['multiplicity_overall'],
        acompleteness_overall = xml_results['acompleteness_overall'],
        amultiplicity_overall = xml_results['amultiplicity_overall'],
        cchalf_overall = xml_results['cchalf_overall'],
        ccanom_overall = xml_results['ccanom_overall'],
        nref_overall = xml_results['nref_overall'],
        nunique_overall = xml_results['nunique_overall'],
        rmerge_inner = xml_results['rmerge_inner'],
        rmeas_inner = xml_results['rmeas_inner'],
        resol_high_inner = xml_results['resol_high_inner'],
        resol_low_inner = xml_results['resol_low_inner'],
        isigma_inner = xml_results['isigma_inner'],
        completeness_inner = xml_results['completeness_inner'],
        multiplicity_inner = xml_results['multiplicity_inner'],
        acompleteness_inner = xml_results['acompleteness_inner'],
        amultiplicity_inner = xml_results['amultiplicity_inner'],
        cchalf_inner = xml_results['cchalf_inner'],
        ccanom_inner = xml_results['ccanom_inner'],
        nref_inner = xml_results['nref_inner'],
        nunique_inner = xml_results['nunique_inner'],
        rmerge_outer = xml_results['rmerge_outer'],
        rmeas_outer = xml_results['rmeas_outer'],
        resol_high_outer = xml_results['resol_high_outer'],
        resol_low_outer = xml_results['resol_low_outer'],
        isigma_outer = xml_results['isigma_outer'],
        completeness_outer = xml_results['completeness_outer'],
        multiplicity_outer = xml_results['multiplicity_outer'],
        acompleteness_outer = xml_results['acompleteness_outer'],
        amultiplicity_outer = xml_results['amultiplicity_outer'],
        cchalf_outer = xml_results['cchalf_outer'],
        ccanom_outer = xml_results['ccanom_outer'],
        nref_outer = xml_results['nref_outer'],
        nunique_outer = xml_results['nunique_outer'],
        refined_beam_x = refined_beam[0],
        refined_beam_y = refined_beam[1],
        filename = os.path.split(start_image)[-1],
        directory = os.path.split(start_image)[0],
    ))



from __future__ import absolute_import, division, print_function

from fast_dp.run_job import run_job

from fast_dp.logger import write

def anomalous_signals(hklin):
    '''
    Compute some measures of anomalous signal: df / f and di / sig(di).
    '''

    from iotbx import mtz

    m = mtz.object(hklin)
    mas = m.as_miller_arrays()

    data = None

    for ma in mas:
        if not ma.anomalous_flag():
            continue
        if str(ma.observation_type()) != 'xray.intensity':
            continue
        data = ma

    if not data:
        raise RuntimeError('no data found')

    df_f = data.anomalous_signal()
    differences = data.anomalous_differences()
    di_sigdi = (sum(abs(differences.data())) /
                sum(differences.sigmas()))

    return df_f, di_sigdi

def merge(hklout='fast_dp.mtz', aimless_log='aimless.log'):
    '''Merge the reflections from XDS_ASCII.HKL with Aimless to get
    statistics - this will use pointless for the reflection file format
    mashing.'''

    run_job('pointless',
            ['-c', 'xdsin', 'XDS_ASCII.HKL', 'hklout', 'xds_sorted.mtz'])

    log = run_job('aimless',
                  ['hklin', 'xds_sorted.mtz', 'hklout', hklout,
                   'xmlout', 'aimless.xml'],
                  ['bins 20', 'run 1 all', 'scales constant',
                   'anomalous on', 'cycles 0', 'output unmerged',
                   'sdcorrection norefine full 1 0 0'])

    fout = open(aimless_log, 'w')

    for record in log:
        fout.write(record)

    fout.close()

    # check for errors

    for record in log:
        if '!!!! No data !!!!' in record:
            raise RuntimeError('aimless complains no data')


    return parse_aimless_log(log)

def parse_aimless_log(log):

    for record in log:
        if 'Low resolution limit  ' in record:
            lres = tuple(map(float, record.split()[-3:]))
        elif 'High resolution limit  ' in record:
            hres = tuple(map(float, record.split()[-3:]))
        elif 'Rmerge  (within I+/I-)  ' in record:
            rmerge = tuple(map(float, record.split()[-3:]))
        elif 'Rmeas (within I+/I-) ' in record:
            rmeas = tuple(map(float, record.split()[-3:]))
        elif 'Mean((I)/sd(I))  ' in record:
            isigma = tuple(map(float, record.split()[-3:]))
        elif 'Completeness  ' in record:
            comp = tuple(map(float, record.split()[-3:]))
        elif 'Multiplicity  ' in record:
            mult = tuple(map(float, record.split()[-3:]))
        elif 'Anomalous completeness  ' in record:
            acomp = tuple(map(float, record.split()[-3:]))
        elif 'Anomalous multiplicity  ' in record:
            amult = tuple(map(float, record.split()[-3:]))
        elif 'Mid-Slope of Anom Normal Probability  ' in record:
            slope = float(record.split()[-3])
        elif 'Total number of observations' in record:
            nref = tuple(map(int, record.split()[-3:]))
        elif 'Total number unique' in record:
            nuniq = tuple(map(int, record.split()[-3:]))
        elif 'DelAnom correlation between half-sets' in record:
            ccanom = tuple(map(float, record.split()[-3:]))
        elif 'Mn(I) half-set correlation CC(1/2)' in record:
            cchalf = tuple(map(float, record.split()[-3:]))

    # copy to internal storage for XML output
    xml_results = {}
    xml_results['rmerge_overall'] = rmerge[0]
    xml_results['rmeas_overall'] = rmeas[0]
    xml_results['resol_high_overall'] = hres[0]
    xml_results['resol_low_overall'] = lres[0]
    xml_results['isigma_overall'] = isigma[0]
    xml_results['completeness_overall'] = comp[0]
    xml_results['multiplicity_overall'] = mult[0]
    xml_results['acompleteness_overall'] = acomp[0]
    xml_results['amultiplicity_overall'] = amult[0]
    xml_results['cchalf_overall'] = cchalf[0]
    xml_results['ccanom_overall'] = ccanom[0]
    xml_results['nref_overall'] = nref[0]
    xml_results['nunique_overall'] = nuniq[0]

    xml_results['rmerge_inner'] = rmerge[1]
    xml_results['rmeas_inner'] = rmeas[1]
    xml_results['resol_high_inner'] = hres[1]
    xml_results['resol_low_inner'] = lres[1]
    xml_results['isigma_inner'] = isigma[1]
    xml_results['completeness_inner'] = comp[1]
    xml_results['multiplicity_inner'] = mult[1]
    xml_results['acompleteness_inner'] = acomp[1]
    xml_results['amultiplicity_inner'] = amult[1]
    xml_results['cchalf_inner'] = cchalf[1]
    xml_results['ccanom_inner'] = ccanom[1]
    xml_results['nref_inner'] = nref[1]
    xml_results['nunique_inner'] = nuniq[1]

    xml_results['rmerge_outer'] = rmerge[2]
    xml_results['rmeas_outer'] = rmeas[2]
    xml_results['resol_high_outer'] = hres[2]
    xml_results['resol_low_outer'] = lres[2]
    xml_results['isigma_outer'] = isigma[2]
    xml_results['completeness_outer'] = comp[2]
    xml_results['multiplicity_outer'] = mult[2]
    xml_results['acompleteness_outer'] = acomp[2]
    xml_results['amultiplicity_outer'] = amult[2]
    xml_results['cchalf_outer'] = cchalf[2]
    xml_results['ccanom_outer'] = ccanom[2]
    xml_results['nref_outer'] = nref[2]
    xml_results['nunique_outer'] = nuniq[2]

    # compute some additional results

    df_f, di_sigdi = anomalous_signals('fast_dp.mtz')
    # print out the results...

    write(80 * '-')

    write('%20s ' % 'Low resolution'     + '%6.2f %6.2f %6.2f' % lres)
    write('%20s ' % 'High resolution'    + '%6.2f %6.2f %6.2f' % hres)
    write('%20s ' % 'Rmerge'             + '%6.3f %6.3f %6.3f' % rmerge)
    write('%20s ' % 'I/sigma'            + '%6.2f %6.2f %6.2f' % isigma)
    write('%20s ' % 'Completeness'       + '%6.1f %6.1f %6.1f' % comp)
    write('%20s ' % 'Multiplicity'       + '%6.1f %6.1f %6.1f' % mult)
    write('%20s ' % 'CC 1/2'             + '%6.3f %6.3f %6.3f' % cchalf)
    write('%20s ' % 'Anom. Completeness' + '%6.1f %6.1f %6.1f' % acomp)
    write('%20s ' % 'Anom. Multiplicity' + '%6.1f %6.1f %6.1f' % amult)
    write('%20s ' % 'Anom. Correlation'  + '%6.3f %6.3f %6.3f' % ccanom)
    write('%20s ' % 'Nrefl'              + '%6d %6d %6d' % nref)
    write('%20s ' % 'Nunique'            + '%6d %6d %6d' % nuniq)
    write('%20s ' % 'Mid-slope'          + '%6.3f' % slope)
    write('%20s ' % 'dF/F'               + '%6.3f' % df_f)
    write('%20s ' % 'dI/sig(dI)'         + '%6.3f' % di_sigdi)

    write(80 * '-')

    return xml_results

if __name__ == '__main__':

    import sys
    parse_aimless_log(open(sys.argv[1]).readlines())



from __future__ import absolute_import, division, print_function

import math
import os
import re
import string

def image2template(filename):
    '''Return a template to match this filename.'''

    # check that the file name doesn't contain anything mysterious
    if filename.count('#'):
        raise RuntimeError('# characters in filename')

    # the patterns in the order I want to test them

    pattern_keys = [r'([^\.]*)\.([0-9]+)\Z',
                    r'(.*)_([0-9]*)\.(.*)',
                    r'(.*?)([0-9]*)\.(.*)']

    # patterns is a dictionary of possible regular expressions with
    # the format strings to put the file name back together

    patterns = {r'([^\.]*)\.([0-9]+)\Z':'%s.%s%s',
                r'(.*)_([0-9]*)\.(.*)':'%s_%s.%s',
                r'(.*?)([0-9]*)\.(.*)':'%s%s.%s'}

    for pattern in pattern_keys:
        match = re.compile(pattern).match(filename)

        if match:
            prefix = match.group(1)
            number = match.group(2)
            try:
                exten = match.group(3)
            except:
                exten = ''

            for digit in string.digits:
                number = number.replace(digit, '#')

            return patterns[pattern] % (prefix, number, exten)

    raise RuntimeError('filename %s not understood as a template' % \
          filename)

def image2image(filename):
    '''Return an integer for the template to match this filename.'''

    # check that the file name doesn't contain anything mysterious
    if filename.count('#'):
        raise RuntimeError('# characters in filename')

    # the patterns in the order I want to test them

    pattern_keys = [r'([^\.]*)\.([0-9]+)\Z',
                    r'(.*)_([0-9]*)\.(.*)',
                    r'(.*?)([0-9]*)\.(.*)']

    for pattern in pattern_keys:
        match = re.compile(pattern).match(filename)

        if match:
            prefix = match.group(1)
            number = match.group(2)
            try:
                exten = match.group(3)
            except:
                exten = ''

            return int(number)

    raise RuntimeError('filename %s not understood as a template' % \
          filename)

def image2template_directory(filename):
    '''Separate out the template and directory from an image name.'''

    directory = os.path.dirname(filename)

    if not directory:

        # then it should be the current working directory
        directory = os.getcwd()

    image = os.path.split(filename)[-1]
    template = image2template(image)

    return template, directory

def find_matching_images(template, directory):
    '''Find images which match the input template in the directory
    provided.'''

    files = os.listdir(directory)

    # to turn the template to a regular expression want to replace
    # however many #'s with EXACTLY the same number of [0-9] tokens,
    # e.g. ### -> ([0-9]{3})

    # change 30/may/2008 - now escape the template in this search to cope with
    # file templates with special characters in them, such as "+" -
    # fix to a problem reported by Joel B.

    length = template.count('#')
    regexp_text = re.escape(template).replace('\\#' * length,
                                              '([0-9]{%d})' % length)
    regexp = re.compile(regexp_text)

    images = []

    for f in files:
        match = regexp.match(f)

        if match:
            images.append(int(match.group(1)))

    images.sort()

    return images

def template_directory_number2image(template, directory, number):
    '''Construct the full path to an image from the template, directory
    and image number.'''

    length = template.count('#')

    # check that the number will fit in the template

    if (math.pow(10, length) - 1) < number:
        raise RuntimeError('number too big for template')

    # construct a format statement to give the number part of the
    # template
    format = '%%0%dd' % length

    # construct the full image name
    image = os.path.join(directory,
                         template.replace('#' * length,
                                          format % number))

    return image



# in preparation for pypi packaging. We are not there yet

from __future__ import absolute_import, division, print_function

from setuptools import find_packages, setup

setup(name='fast_dp',
      description='fast_dp',
      author='Diamond Light Source',
      author_email='scientificsoftware@diamond.ac.uk',

      version='0.4',

      url='https://github.com/DiamondLightSource/fast_dp',
      download_url="https://github.com/DiamondLightSource/fast_dp/releases",
      license='Apache-2.0',

      install_requires=[],
      packages=['fast_dp'],
      package_data={'fast_dp': ['templates/*.INP', 'templates/ispyb.xml']},

      entry_points={
        'libtbx.dispatcher.script': [
          'fast_dp=fast_dp',
          'fast_rdp=fast_rdp',
          'header2edna_xml=header2edna_xml',
        ],
      },
      scripts=[
        'bin/fast_dp',
        'bin/fast_rdp',
        'bin/header2edna_xml',
      ],

      tests_require=['mock', 'procrunner', 'pytest'],
      zip_safe=False,
      classifiers = [
        'Development Status :: 4 - Beta',
#       'License :: OSI Approved :: Apache Software License 2.0 (Apache-2.0)', # eventually. https://github.com/pypa/warehouse/issues/2996
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX :: Linux',
     ],
)



from __future__ import absolute_import, division, print_function

import os
import subprocess

def run_job(executable, arguments = [], stdin = [], working_directory = None):
    '''Run a program with some command-line arguments and some input,
    then return the standard output when it is finished.'''

    if working_directory is None:
        working_directory = os.getcwd()

    command_line = '%s' % executable
    for arg in arguments:
        command_line += ' "%s"' % arg

    popen = subprocess.Popen(command_line,
                             bufsize = 1,
                             stdin = subprocess.PIPE,
                             stdout = subprocess.PIPE,
                             stderr = subprocess.STDOUT,
                             cwd = working_directory,
                             universal_newlines = True,
                             shell = True)

    for record in stdin:
        popen.stdin.write('%s\n' % record)

    popen.stdin.close()

    output = []

    while True:
        record = popen.stdout.readline()
        if not record:
            break

        output.append(record)

    return output

def get_number_cpus():
    '''Portably get the number of processor cores available.'''

    # Windows NT derived platforms

    if os.name == 'nt':
        return int(os.environ['NUMBER_OF_PROCESSORS'])

    # linux

    if os.path.exists('/proc/cpuinfo'):
        n_cpu = 0

        for record in open('/proc/cpuinfo', 'r').readlines():
            if not record.strip():
                continue
            if 'processor' in record.split()[0]:
                n_cpu += 1

        return n_cpu

    # os X

    output = subprocess.Popen(['system_profiler', 'SPHardwareDataType'],
                              stdout = subprocess.PIPE).communicate()[0]

    ht = 1

    for record in output.split('\n'):
        if 'Intel Core i7' in record:
            ht = 2
        if 'Total Number Of Cores' in record:
            return ht * int(record.split()[-1])
        if 'Total Number of Cores' in record:
            return ht * int(record.split()[-1])

    return -1

if __name__ == '__main__':
    os.environ['FAST_DP_FORKINTEGRATE'] = '1'
    print(''.join(run_job('env')))



from __future__ import absolute_import, division, print_function

import os
import procrunner
import pytest

def test_fast_dp_X4_wide(tmpdir):
  X4_wide = '/dls/science/groups/scisoft/DIALS/regression_data/X4_wide'
  if not os.path.exists(X4_wide):
    pytest.skip('Could not find test image file')

  image = os.path.join(X4_wide, 'X4_wide_M1S4_2_0001.cbf')

  bin = os.path.split(__file__)[0].replace('src', 'bin')
  fast_dp = os.path.join(bin, 'fast_dp')

  cmd = [ fast_dp, '-a', 'Ba', image ]
  with tmpdir.as_cwd():
    result = procrunner.run(cmd)

  assert result['stderr'] == '', 'fast_dp output to STDERR:'
  assert result['exitcode'] == 0, 'fast_dp non-zero exit code'
  for output in ['fast_dp.mtz', 'fast_dp.log']:
    assert tmpdir.join(output).check(), 'No output found'
